/*
 * tclock.c	Tclock creates a "transparent" clock for the X Window System
 *		using the shaped-window extension of X11R4.
 *
 * Copyright (C) 1991 Rob Mayoff.
 *
 * E-mail: mayoff@ccwf.cc.utexas.edu
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 1, or (at your option)
 * any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */ 

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <math.h>
#include <time.h>
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xresource.h>
#include <X11/extensions/shape.h>
#include <signal.h>
#include <errno.h>
#include <ctype.h>
#include "defaults.h"

#define HOURHAND	0
#define MINUTEHAND	1
#define SECONDHAND	2
#define NUMHANDS	3

#define NUMVERTICES	5

typedef struct
{
  XPoint	vertex[NUMVERTICES];
  int		offset[NUMVERTICES];
  int		radius[NUMVERTICES];
  int		position;
}	Hand;

Display *display;
int     screen_number;
Window  rootWindow, mainWindow;
int     ShapeEventNumber, ShapeErrorNumber;
time_t  lclock;
struct tm lastTimes, times;
Pixmap  border, clip;
GC      clearGC, drawGC;
double  cos60[60], sin60[60];
Hand	hand[NUMHANDS];
XArc    tick[12];
XrmDatabase	db;
int	clockRadius;
int	tickRadius;
int	tickSize;
double	scale;
char	appName[200], appClass[200];
XSizeHints	mainSize, savedHints;

char *
getOption(o)
  char *o;
{
  XrmValue	value;
  char	*type, key[200], keyClass[200], *p;

  
  /* Make the main key */
  sprintf(key, "%s.%s", appName, o);

  /* Make the class key */
  sprintf(keyClass, "%s.%s", appClass, o);
  for(p = keyClass; *p; p++)
    if(*p == '.' && *(p+1))
      *(p+1) = toupper(*(p+1));

  if(XrmGetResource(db, key, keyClass, &type, &value))
    return value.addr;
  else
    return 0;
}

void
initHand(h, s)
  int h;		/* Hand to initialize */
  char *s;		/* Name of hand */
{
  char	rn[1000];	/* Resource name */
  static
  char	*vn[NUMVERTICES-1] =	/* Names of vertices */
    { "tip", "right", "middle", "left" };
  int	i;

  clockRadius = -1;
  for(i = 0; i < NUMVERTICES-1; i++)
  {
    sprintf(rn, "%s.%s.radius", s, vn[i]);
    hand[h].radius[i] = scale * atoi(getOption(rn)) + 0.5;
    if(clockRadius < hand[h].radius[i])
      clockRadius = hand[h].radius[i];
    sprintf(rn, "%s.%s.offset", s, vn[i]);
    hand[h].offset[i] = atoi(getOption(rn));
  }

  hand[h].radius[i] = hand[h].radius[0];
  hand[h].offset[i] = hand[h].offset[0];
}

void
initMisc()
{
  int i, j;
  extern
  double	atof();

  for(i=0; i<60; i++)
  {
    cos60[i] = cos((double)i * M_PI * 2.0 / 60);
    sin60[i] = sin((double)i * M_PI * 2.0 / 60);
  }

  initHand(HOURHAND, "hour");
  initHand(MINUTEHAND, "minute");
  initHand(SECONDHAND, "second");

  tickRadius = scale * atoi(getOption("tick.radius")) + 0.5;
  tickSize = scale * atoi(getOption("tick.size")) + 0.5;
  if((tickRadius + tickSize) > clockRadius)
    clockRadius = tickRadius+tickSize;

  for(i=0; i<12; i++)
  {
    j = (i + 9) % 12;
    tick[i].x =
      0.5 + (double)clockRadius + (double)cos60[j*5] * tickRadius - tickSize;
    tick[i].y =
      0.5 + (double)clockRadius + (double)sin60[j*5] * tickRadius - tickSize;
    tick[i].width = tick[i].height = tickSize*2;
    tick[i].angle1 = 0;
    tick[i].angle2 = 360*64;
  }
}

void drawHand(h, p)
  int h;		/* Hand number to draw */
  int p;		/* Position at which to draw it */
{
  int i, j;

  if(p != hand[h].position)
  {
    hand[h].position = p;

    for(j = 0; j < NUMVERTICES - 1; j++)
    {
      i = (p + 45 + hand[h].offset[j]);
      while(i >= 60)	i-=60;
      hand[h].vertex[j].x =
	hand[h].radius[j] * cos60[i] + 0.5 + clockRadius;
      hand[h].vertex[j].y =
	hand[h].radius[j] * sin60[i] + 0.5 + clockRadius;
    }

    /* Close the polygon for XDrawLines (sigh) */
    hand[h].vertex[j] = hand[h].vertex[0];
  }

  XFillPolygon(display, clip, drawGC, hand[h].vertex, NUMVERTICES-1,
    Nonconvex, CoordModeOrigin);
  XDrawLines(display, clip, clearGC, hand[h].vertex, NUMVERTICES,
    CoordModeOrigin);

  XFillPolygon(display, border, drawGC, hand[h].vertex, NUMVERTICES-1,
    Nonconvex, CoordModeOrigin);
  XDrawLines(display, border, drawGC, hand[h].vertex, NUMVERTICES,
    CoordModeOrigin);
}

void
reshape()
{
  XFillRectangle(display, border, clearGC, 0, 0,
    clockRadius*2+1, clockRadius*2+1);
  XFillRectangle(display, clip, clearGC, 0, 0,
    clockRadius*2+1, clockRadius*2+1);

  if(tickRadius > 0)
  {
    XFillArcs(display, clip, drawGC, tick, 12);
    XDrawArcs(display, clip, clearGC, tick, 12);

    XFillArcs(display, border, drawGC, tick, 12);
    XDrawArcs(display, border, drawGC, tick, 12);
  }

  if(hand[SECONDHAND].radius[0])
    drawHand(SECONDHAND, times.tm_sec);
  if(hand[MINUTEHAND].radius[0])
    drawHand(MINUTEHAND, times.tm_min);
  if(hand[HOURHAND].radius[0])
    drawHand(HOURHAND, times.tm_hour);	/* Hour converted in sigalrm() */

  XShapeCombineMask(display, mainWindow, ShapeBounding, 0, 0, border, ShapeSet);
  XShapeCombineMask(display, mainWindow, ShapeClip, 0, 0, clip, ShapeSet);

  XFlush(display);

  /* This call to XPending ensures that if the window gets closed, we die */
  XPending(display);
}

XTextProperty *
strToTP(s)
  char *s;
{

  XTextProperty *tp = (XTextProperty *)malloc(sizeof *tp);
  XStringListToTextProperty(&s, 1, tp);
  return tp;
}

int
sigalrm()
{
  lclock = time(0);
  times = *localtime(&lclock);
  times.tm_hour = times.tm_hour * 5 + times.tm_min / 12;
  reshape();
  return 0;
}

unsigned long
getColor(c)
  char *c;
{
  XColor	color;

  if(!XParseColor(display, DefaultColormap(display, screen_number),
    c, &color))
    return 0;
  
  XAllocColor(display, DefaultColormap(display, screen_number), &color);

  return color.pixel;
}

void
usage(name)
  char *name;
{
  fprintf(stderr, "usage: %s [options]\n", name);
  fprintf(stderr, "Options: (options followed by resource equivalents)\n");
  fprintf(stderr, "  -background      .background       Background color\n");
  fprintf(stderr, "  -bg              .background       ...\n");
  fprintf(stderr, "  -border          .border           Border color\n");
  fprintf(stderr, "  -borderWidth     .borderWidth      Border width\n");
  fprintf(stderr, "  -display         .display          X server name\n");
  fprintf(stderr, "  -geometry        .geometry		Clock location\n");
  fprintf(stderr, "  -scale           .scale            Scaling factor\n");
  fprintf(stderr, "  -tickradius      .tick.radius      Tick distance\n");
  fprintf(stderr, "  -ticksize        .tick.size        Tick size\n");
  fprintf(stderr, "  -width           .borderWidth      Border width\n");
  fprintf(stderr, "  -xrm             other resources\n");
  fprintf(stderr, "See defaults.h for other resources.  A tick radius of\n");
  fprintf(stderr, "zero will disable tick marks, for those who like a very\n");
  fprintf(stderr, "abstract clock.  The size part of the geometry strin\n");
  fprintf(stderr, "is ignored - use -scale to adjust the clock size.\n");
  fprintf(stderr, "\nTclock version 1.0, Copyright (C) 1991 Rob Mayoff.\n");
  fprintf(stderr, "Tclock comes with ABSOLUTELY NO WARRANTY.  This is free software, and\n");
  fprintf(stderr, "you are welcome to redistribute it under certain conditions; see the file\n");
  fprintf(stderr, "COPYING included with the distribution for details.\n");
}

void
makeAppName(r)
  char *r;
{
  char *p;

  for(p = r+strlen(r); p >= r && *p != '/'; p--)
    ;
  strcpy(appName, ++p);
  strcpy(appClass, "Tclock");
}

int 
main(ac, av)
  int ac;
  char **av;
{
  XSetWindowAttributes  mainAtts;
  XWMHints      mainWM;
  XClassHint    mainClass;
  struct itimerval      old, new;
  int           sigalrm();
  XGCValues     gcv;
  XrmDatabase	cmdDB;
  extern
  char		*getenv();
  char		*displayName, fn[1000];
  char		*sdb;
  int		parseReturn;
  sigset_t	mask;

  XrmParseCommand(&db, option, numOptions, "tclock", &ac, av);
  if(ac>1)
  {
    usage(av[0]);
    exit(1);
  }

  makeAppName(av[0]);

  displayName = getOption("display");
  cmdDB = db;

  if(!displayName)
    displayName = getenv("DISPLAY");
  if(!displayName)
    displayName = ":0";
  display = XOpenDisplay(displayName);
  if(display == 0)
  {
    fprintf(stderr, "Couldn't open display %s\n", displayName);
    exit(-1);
  }

  screen_number = DefaultScreen(display);
  rootWindow = RootWindow(display, screen_number);

  if(!XShapeQueryExtension(display, &ShapeEventNumber, &ShapeErrorNumber))
    exit(1);

  db = XrmGetStringDatabase(tclockDefaults);
  if(sdb = XResourceManagerString(display))
    XrmMergeDatabases(XrmGetStringDatabase(sdb), &db);
  
  sprintf(fn, "%s/.Xdefaults", getenv("HOME"));
  if(access(fn, R_OK) == 0)
    XrmMergeDatabases(XrmGetFileDatabase(fn), &db);

  if(getenv("XENVIRONMENT") && access(getenv("XENVIRONMENT"), R_OK) == 0)
    XrmMergeDatabases(XrmGetFileDatabase(getenv("XENVIRONMENT")), &db);

  XrmMergeDatabases(cmdDB, &db);

  scale = atof(getOption("scale"));
  initMisc();
  
  if(getOption("geometry"))
    parseReturn = XParseGeometry(getOption("geometry"), &(mainSize.x),
      &(mainSize.y), &(mainSize.width), &(mainSize.height));
  else
    parseReturn = NoValue;

  mainSize.width = clockRadius*2+1;
  mainSize.height = clockRadius*2+1;
  mainSize.flags = 0;
  if(parseReturn&(XValue|YValue))
  {
    mainSize.flags = USPosition;
    if(parseReturn&XValue && parseReturn&XNegative)
      mainSize.x =
	DisplayWidth(display, screen_number) - mainSize.width + mainSize.x-1;
    if(parseReturn&YValue && parseReturn&YNegative)
      mainSize.y =
	DisplayHeight(display, screen_number) - mainSize.height + mainSize.y-1;
  }

  mainAtts.border_pixel = getColor(getOption("border"));
  mainAtts.background_pixel = getColor(getOption("background"));
  mainAtts.save_under = 1;
  mainAtts.event_mask = StructureNotifyMask|ButtonPressMask|ButtonReleaseMask;
  mainAtts.do_not_propagate_mask = 0;
  mainWindow = XCreateWindow(display, rootWindow,
    mainSize.x, mainSize.y, mainSize.width, mainSize.height,
    0, DefaultDepth(display, screen_number), InputOutput,
    DefaultVisual(display, screen_number),
    CWBorderPixel|CWBackPixel|CWSaveUnder|CWEventMask|CWDontPropagate,
      &mainAtts);
  
  savedHints = mainSize;

  mainWM.flags = 0;
  mainClass.res_name = mainClass.res_class = "xclock";
  XSetWMProperties(display, mainWindow, strToTP("xclock"), strToTP("xclock"),
    av, ac, &mainSize, &mainWM, &mainClass);

  border = XCreatePixmap(display, mainWindow, clockRadius*2+1, clockRadius*2+1, 1);
  clip = XCreatePixmap(display, mainWindow, clockRadius*2+1, clockRadius*2+1, 1);

  gcv.function = GXclear;
  gcv.join_style = JoinMiter;
  gcv.line_width = atoi(getOption("borderWidth"));
  clearGC = XCreateGC(display, border,
    GCFunction|GCJoinStyle|GCLineWidth, &gcv);
  gcv.function = GXset;
  drawGC = XCreateGC(display, border,
    GCFunction|GCJoinStyle|GCLineWidth, &gcv);

  sigalrm();
  XMapRaised(display, mainWindow);
  XFlush(display);

  signal(SIGALRM, (void *)sigalrm);
  new.it_interval.tv_sec = 1;
  new.it_value.tv_sec = 1;
  new.it_interval.tv_usec = 0;
  new.it_value.tv_usec = 0;
  setitimer(ITIMER_REAL, &new, &old);

  sigemptyset(&mask);
  for(;;)
    sigsuspend(&mask);
  
  return 0;
}
